"""
trio
----

Trio is also supported by cyares it works the same way as the `cyares.aio` module
and **aiodns** so it should not be too tricky to navigate.

"""

from __future__ import annotations

import socket
from concurrent.futures import Future as ccFuture
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Literal,
    Optional,
    Sequence,
    TypeVar,
    overload,
)

import trio
from deprecated_params import deprecated_params
from trio.lowlevel import current_clock, current_trio_token

from .channel import *
from .resulttypes import *

try:
    _wait_readable = trio.lowlevel.wait_readable
    _wait_writable = trio.lowlevel.wait_writable
except AttributeError:
    _wait_readable = trio.lowlevel.wait_socket_readable
    _wait_writable = trio.lowlevel.wait_socket_writable


class CancelledError(Exception):
    pass


class Timer:
    def __init__(self, cb: Callable[..., None], timeout: Optional[float] = None):
        self.clock = current_clock()
        self.deadline = self.clock.current_time() + (timeout or 1)
        self.cb = cb
        self._running = False
        self._close_event: Optional[trio.Event] = None

    def _check_soon(self):
        current_trio_token().run_sync_soon(self._check_time)

    def _check_time(self):
        if self._close_event:
            # Timer was turned off by the end user to prepare for closure
            return self._close_event.set()

        if self.clock.current_time() >= self.deadline:
            self.cb()
            self._running = False
        else:
            self._check_soon()

    def start(self):
        self._running = True
        self._check_soon()

    def cancel(self):
        if not self._close_event:
            self._close_event = trio.Event()

    async def close(self) -> None:
        """Waits for the timer to come to a callback where it can shut down"""
        if not self._running:
            return
        return await self._close_event.wait()


query_type_map = {
    "A": QUERY_TYPE_A,
    "AAAA": QUERY_TYPE_AAAA,
    "ANY": QUERY_TYPE_ANY,
    "CAA": QUERY_TYPE_CAA,
    "CNAME": QUERY_TYPE_CNAME,
    "MX": QUERY_TYPE_MX,
    "NAPTR": QUERY_TYPE_NAPTR,
    "NS": QUERY_TYPE_NS,
    "PTR": QUERY_TYPE_PTR,
    "SOA": QUERY_TYPE_SOA,
    "SRV": QUERY_TYPE_SRV,
    "TXT": QUERY_TYPE_TXT,
}

query_class_map = {
    "IN": QUERY_CLASS_IN,
    "CHAOS": QUERY_CLASS_CHAOS,
    "HS": QUERY_CLASS_HS,
    "NONE": QUERY_CLASS_NONE,
    "ANY": QUERY_CLASS_ANY,
}

_T = TypeVar("_T")


class Future(Generic[_T]):
    __class_getitem__ = classmethod(GenericAlias)

    def __init__(self, fut: ccFuture[_T], uses_thread: bool = True):
        self._exc = None
        self._result = None
        self._cancelled = False
        self.event = trio.Event()
        self._callbacks: list[Callable[["Future[_T]"], None]] = []
        # Token will allow use to reach the homethread allowing for a seamless transition
        self._token = current_trio_token()
        # all we needed the other future for was preparing to chain it.
        fut.add_done_callback(self.__on_done)
        # determines if were in the Home Thread
        self._uses_thread = uses_thread

    def __execute_callbacks(self):
        for cb in self._callbacks:
            cb(self)
        self._callbacks.clear()

    def __handle_done(self):
        self.event.set()
        self.__execute_callbacks()

    def __on_done(self, fut: ccFuture[_T]) -> None:
        if fut.cancelled():
            self._cancelled = True
        elif exc := fut.exception():
            self._exc = exc
        else:
            self._result = fut.result()

        # We need to call everything else from the home-thread to prevent any errors...
        if self._uses_thread:
            trio.from_thread.run_sync(self.__handle_done, trio_token=self._token)
        else:
            self.__handle_done()

    def done(self) -> bool:
        return self.event.is_set()

    def cancel(self):
        if not self._cancelled and not self.done():
            self._cancelled = True

    def cancelled(self) -> bool:
        return self.event.is_set() and self._cancelled

    async def _wait(self) -> _T:
        await self.event.wait()
        if self._cancelled:
            raise CancelledError()
        elif self._exc:
            raise self._exc
        return self._result  # type: ignore

    def __await__(self):
        return self._wait().__await__()


@deprecated_params(
    ["sock_state_cb"],
    "attempting to pass socket_state_cb will throw an Exception instead "
    "of being ignored in a future version of cyares",
    removed_in="0.1.8",
)
class DNSResolver:
    def __init__(
        self,
        servers: list[str] | None = None,
        event_thread: bool = False,  # turned off by default but you can always pass it if you wish...
        timeout: float | None = None,
        flags: int | None = None,
        tries: int | None = None,
        ndots: object | None = None,
        tcp_port: int | None = None,
        udp_port: int | None = None,
        domains: list[str] | None = None,
        lookups: str | bytes | bytearray | memoryview[int] | None = None,
        socket_send_buffer_size: int | None = None,
        socket_receive_buffer_size: int | None = None,
        rotate: bool = False,
        local_ip: str | bytes | bytearray | memoryview[int] | None = None,
        local_dev: str | bytes | bytearray | memoryview[int] | None = None,
        resolvconf_path=None,
        
        **kw
    ):
        kw.pop("socket_state_cb", None)

        self._channel = Channel(
            servers=servers,
            event_thread=event_thread,
            sock_state_cb=self._socket_state_cb if not event_thread else None,
            flags=flags,
            tries=tries,
            ndots=ndots,
            tcp_port=tcp_port,
            udp_port=udp_port,
            domains=domains,
            lookups=lookups,
            socket_send_buffer_size=socket_send_buffer_size,
            socket_receive_buffer_size=socket_receive_buffer_size,
            rotate=rotate,
            local_ip=local_ip,
            local_dev=local_dev,
            resolvconf_path=resolvconf_path,
            **kw,
        )
        self._manager = trio.open_nursery()
        self._nursery: Optional[trio.Nursery] = None
        self._read_fds: set[int] = set()
        self._write_fds: set[int] = set()
        self._timeout = kw.get("timeout", None)
        self._timer = None
        self._token = current_trio_token()

    async def __aenter__(self):
        self._nursery = await self._manager.__aenter__()
        return self

    async def __aexit__(self, *args):
        await self._manager.__aexit__(*args)

    async def _handle_read(self, fd: int):
        while fd in self._read_fds:
            try:
                await _wait_readable(fd)
                self._channel.process_read_fd(fd)
            except OSError:
                if fd not in self._read_fds:
                    break
                raise

    async def _handle_write(self, fd: int):
        while fd in self._write_fds:
            try:
                await _wait_writable(fd)
                self._channel.process_write_fd(fd)
            except OSError:
                if fd not in self._write_fds:
                    break
                raise

    def _timer_cb(self) -> None:
        if self._read_fds or self._write_fds:
            self._channel.process_fd(CYARES_SOCKET_BAD, CYARES_SOCKET_BAD)
            self._start_timer()
        else:
            timer = self._timer
            # Timer Cleanup
            self._nursery.start_soon(timer.close, name=f"Timer Cleanup: {self!r}")
            self._timer = None

    def _start_timer(self) -> None:
        timeout = self._timeout
        if timeout is None or timeout < 0 or timeout > 1:
            timeout = 1
        elif timeout == 0:
            timeout = 0.1

        self._timer = Timer(self._timer_cb, self._timeout)

    def _socket_state_cb(self, fd: int, read: bool, write: bool) -> None:
        if read or write:
            if read:
                self._read_fds.add(fd)
                self._nursery.start_soon(self._handle_read, fd)

            if write:
                self._write_fds.add(fd)
                self._nursery.start_soon(self._handle_write, fd)

            if self._timer is None:
                self._start_timer()
        else:
            # socket is now closed
            if fd in self._read_fds:
                self._read_fds.discard(fd)

            if fd in self._write_fds:
                self._write_fds.discard(fd)

            if not self._read_fds and not self._write_fds and self._timer is not None:
                self._timer.cancel()
                self._timer = None

    @property
    def nameservers(self) -> Sequence[str]:
        return self._channel.servers

    @nameservers.setter
    def nameservers(self, value: Iterable[str | bytes]) -> None:
        self._channel.servers = value

    @overload
    def query(
        self, host: str, qtype: Literal["A"], qclass: str | None = ...
    ) -> Future[list[ares_query_a_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["AAAA"], qclass: str | None = ...
    ) -> Future[list[ares_query_aaaa_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["CAA"], qclass: str | None = ...
    ) -> Future[list[ares_query_caa_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["CNAME"], qclass: str | None = ...
    ) -> Future[ares_query_cname_result]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["MX"], qclass: str | None = ...
    ) -> Future[list[ares_query_mx_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["NAPTR"], qclass: str | None = ...
    ) -> Future[list[ares_query_naptr_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["NS"], qclass: str | None = ...
    ) -> Future[list[ares_query_ns_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["PTR"], qclass: str | None = ...
    ) -> Future[list[ares_query_ptr_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["SOA"], qclass: str | None = ...
    ) -> Future[ares_query_soa_result]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["SRV"], qclass: str | None = ...
    ) -> Future[list[ares_query_srv_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["TXT"], qclass: str | None = ...
    ) -> Future[list[ares_query_txt_result]]: ...

    def query(
        self, host: str, qtype: str, qclass: str | None = None
    ) -> Future[list[Any]] | Future[Any]:
        try:
            qtype = query_type_map[qtype]
        except KeyError as e:
            raise ValueError(f"invalid query type: {qtype}") from e
        if qclass is not None:
            try:
                qclass = query_class_map[qclass]
            except KeyError as e:
                raise ValueError(f"invalid query class: {qclass}") from e

        # we use a different technique than pycares to try and
        # aggressively prevent vulnerabilities

        return self._wrap_future(self._channel.query(host, qtype, qclass))

    def gethostbyname(
        self, host: str, family: socket.AddressFamily
    ) -> Future[ares_host_result]:
        return self._wrap_future(self._channel.gethostbyname(host, family))

    def getaddrinfo(
        self,
        host: str,
        family: socket.AddressFamily = socket.AF_UNSPEC,
        port: int | None = None,
        proto: int = 0,
        type: int = 0,
        flags: int = 0,
    ) -> Future[ares_addrinfo_result]:
        return self._wrap_future(
            self._channel.getaddrinfo(
                host, port, family=family, type=type, proto=proto, flags=flags
            )
        )

    def gethostbyaddr(
        self, name: str | bytes | bytearray | memoryview
    ) -> Future[ares_host_result]:
        return self._wrap_future(self._channel.gethostbyaddr(name))

    async def close(self) -> None:
        """
        Cleanly close the DNS resolver.

        This should be called to ensure all resources are properly released.
        After calling close(), the resolver should not be used again.
        """
        await self._cleanup()

    def getnameinfo(
        self,
        sockaddr: tuple[str, int] | tuple[str, int, int, int],
        flags: int = 0,
    ) -> Future[ares_nameinfo_result]:
        return self._wrap_future(self._channel.getnameinfo(sockaddr, flags))

    def _wrap_future(self, fut: ccFuture[_T]) -> Future[_T]:
        # use the event_thread readonly property to determine if we will be in the home thread or not.
        return Future(fut, self._channel.event_thread)

    def cancel(self) -> None:
        """Cancels all running futures queued by this dns resolver"""
        self._channel.cancel()

    async def _cleanup(self) -> None:
        """Cleanup timers and file descriptors when closing resolver."""
        if self._closed:
            return
        # Mark as closed first to prevent double cleanup
        self._closed = True
        # Cancel timer if running
        if self._timer is not None:
            # perform safe trio cleanup
            await self._timer.close()
            self._timer = None

        self._read_fds.clear()
        self._write_fds.clear()
        self._channel.cancel()



import os
import sys
from concurrent.futures import Future

import pycares
import pytest
from dnslib.dns import RR, DNSQuestion
from dnslib.server import BaseResolver, DNSHandler, DNSRecord, DNSServer
from pytest_codspeed import BenchmarkFixture

import cyares
from cyares.exception import AresError

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class DummyServer(DNSServer):
    def __enter__(self) -> Self:
        self.start_thread()
        return self

    def __exit__(self, *args) -> None:
        self.stop()


class DummyResolver(BaseResolver):
    def __init__(self):
        self.answers = {
            # TODO: More Query Types Like MX & AAAA for now I'm just benchmarking A
            "example1.com.": RR.fromZone("example1. 60 A 127.0.0.1"),
            "example2.com.": RR.fromZone("example2. 60 A 127.0.0.2"),
            "example3.com.": RR.fromZone("example3. 60 A 127.0.0.3"),
        }

    def resolve(self, request: DNSRecord, handler: DNSHandler):
        reply = request.reply()
        for q in request.questions:
            reply.add_answer(*self.answers[str(q.qname)])
        return reply


@pytest.fixture(scope="function", params=("tcp", "udp"), ids=str)
def dns_server_protocol(request: pytest.FixtureRequest):
    with DummyServer(DummyResolver(), "localhost", tcp=request.param == "tcp"):
        yield f"{request.param}://127.0.0.1:53"


@pytest.fixture(scope="function")
def dns_server_udp_only(request: pytest.FixtureRequest):
    with DummyServer(DummyResolver(), "localhost"):
        yield


@pytest.fixture(scope="session")
def cyares_channel() -> type[cyares.Channel]:
    return cyares.Channel


@pytest.fixture(scope="session")
def pycares_channel() -> type[pycares.Channel]:
    return pycares.Channel


@pytest.fixture(scope="session")
def one_hundred_queries():
    items = ["example1.com", "example2.com", "example3.com"] * 33
    items.append("example1.com")
    return items


def pycares_query(channel: pycares.Channel, query: str) -> Future:
    fut = Future()

    def on_result(result, status: int):
        nonlocal fut
        if status:
            fut.set_exception(AresError(status))
        fut.set_result(result)

    channel.query(query, pycares.QUERY_TYPE_A, on_result)
    return fut


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS, reason="Github Doesn't like running a DNS Dummy Server"
)
def test_pycares_100_queries_benchmark(
    one_hundred_queries: list[str],
    dns_server_udp_only,
    benchmark: BenchmarkFixture,
    pycares_channel,
) -> None:
    channel = pycares_channel(servers=["127.0.0.1"], event_thread=True)

    def setup():
        return (one_hundred_queries,), {}

    def target(queries: list[str]):
        for query in queries:
            pycares_query(channel, query).result(10)

    def teardown(data):
        channel.cancel()

    benchmark.pedantic(
        target, setup=setup, teardown=teardown, rounds=100, warmup_rounds=10
    )


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS, reason="Github Doesn't like running a DNS Dummy Server"
)
def test_cyares_100_queries_benchmark(
    one_hundred_queries: list[str],
    dns_server_udp_only,
    benchmark: BenchmarkFixture,
) -> None:
    channel = cyares.Channel(servers=["127.0.0.1"], event_thread=True)

    def setup():
        return (one_hundred_queries,), {}

    def target(queries: list[str]):
        for query in queries:
            channel.query(query, "A").result(10)

    def teardown(data):
        channel.cancel()

    benchmark.pedantic(
        target, setup=setup, teardown=teardown, rounds=100, warmup_rounds=10
    )


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS, reason="Github Doesn't like running a DNS Dummy Server"
)
def test_pycares_100_queries_benchmark_concurrently(
    one_hundred_queries: list[str],
    dns_server_udp_only,
    benchmark: BenchmarkFixture,
    pycares_channel,
) -> None:
    channel = pycares_channel(servers=["127.0.0.1"], event_thread=True)

    def setup():
        return (one_hundred_queries,), {}

    def target(queries: list[str]):
        for f in [pycares_query(channel, query) for query in queries]:
            f.result(10)

    def teardown(data):
        channel.cancel()

    benchmark.pedantic(
        target, setup=setup, teardown=teardown, rounds=100, warmup_rounds=10
    )


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS, reason="Github Doesn't like running a DNS Dummy Server"
)
def test_cyares_100_queries_benchmark_concurrently(
    one_hundred_queries: list[str], dns_server_udp_only, benchmark: BenchmarkFixture
) -> None:
    channel = cyares.Channel(servers=["127.0.0.1"], event_thread=True)

    def setup():
        return (one_hundred_queries,), {}

    def target(queries: list[str]):
        for f in [channel.query(query, "A") for query in queries]:
            f.result(10)

    def teardown(data):
        channel.cancel()

    benchmark.pedantic(
        target, setup=setup, teardown=teardown, rounds=100, warmup_rounds=10
    )



from __future__ import annotations

import ipaddress
import re
import socket
import sys
from typing import Any, Callable, Generator

import pytest

from select import select

from cyares import Channel
from cyares.channel import CYARES_SOCKET_BAD
from cyares.exception import AresError

ChannelType = Callable[..., Channel]


# TODO: To test getsock() in a future update I'll bring in
# the pycares workflow and see what kinds of tweaks can be
# made to it.

# TODO: add pycares workflow in 0.1.2 for more aggressive stress testing


def wait(channel: Channel):
    while True:
        read_fds, write_fds = channel.getsock()
        if not read_fds and not write_fds:
            break
        timeout = channel.timeout()
        if timeout == 0.0:
            channel.process_fd(CYARES_SOCKET_BAD, CYARES_SOCKET_BAD)
            continue
        rlist, wlist, _ = select(read_fds, write_fds, [], timeout)
        for fd in rlist:
            channel.process_read_fd(fd)
        for fd in wlist:
            channel.process_write_fd(fd)


@pytest.fixture(scope="function")
def c() -> Generator[Any, Any, tuple[Channel, bool]]:
    # should be supported on all operating systems...

    with Channel(
        servers=[
            # Added more dns servers incase we lag behind or one kicks us off.
            # unfortunately there's no way for me to contact and say hi, I'm writing
            # a new dns resolver can I use your server for stress-testing?
            # Maybe in the future we can make a local dns server to stress test our things.
            "8.8.8.8",
            "8.8.4.4",
            "1.0.0.1",
            "1.1.1.1",
            "141.1.27.249",
            "194.190.225.2",
            "194.225.16.5",
            "91.185.6.10",
            "194.2.0.50",
            "66.187.16.5",
            "83.222.161.130",
            "69.60.160.196",
            "194.150.118.3",
            "84.8.2.11",
            "195.175.39.40",
            "193.239.159.37",
            "205.152.6.20",
            "82.151.90.1",
            "144.76.202.253",
            "103.3.46.254",
            "5.144.17.119",
        ],
        event_thread=True,
    ) as channel:
        yield channel


def test_open_and_closure() -> None:
    with Channel() as c:
        pass


def test_nameservers() -> None:
    with Channel(servers=["8.8.8.8", "8.8.4.4"]) as c:
        assert c.servers == ["8.8.8.8:53", "8.8.4.4:53"]


def test_mx_dns_query(c: Channel) -> None:
    fut = c.query("gmail.com", query_type="MX")


def test_a_dns_query(c: Channel) -> None:
    for f in [
        c.query("google.com", "A"),
        c.query("llhttp.org", "A"),
        c.query("llparse.org", "A"),
    ]:
        assert f.result()


def test_cancelling(c: Channel) -> None:
    for f in [
        c.query("google.com", "A"),
        c.query("llhttp.org", "A"),
        c.query("llparse.org", "A"),
    ]:
        f.cancel()


def test_a_dns_query_fail(c: Channel) -> None:
    with pytest.raises(
        AresError,
        match=re.escape("[ARES_ENODATA : 1] DNS server returned answer with no data"),
    ):
        c.query("hgf8g2od29hdohid.com", "A").result()


def test_query_aaaa(c: Channel) -> None:
    assert c.query("ipv6.google.com", "AAAA").result()


def test_query_cname(c: Channel) -> None:
    assert c.query("www.amazon.com", "CNAME").result()


def test_query_mx(c: Channel) -> None:
    assert c.query("google.com", "MX").result()


def test_query_ns(c: Channel) -> None:
    assert c.query("google.com", "NS").result()


def test_query_txt(c: Channel) -> None:
    assert c.query("google.com", "TXT").result()


def test_query_soa(c: Channel) -> None:
    assert c.query("google.com", "SOA").result()


def test_query_srv(c: Channel) -> None:
    assert c.query("_xmpp-server._tcp.jabber.org", "SRV").result()


def test_query_naptr(c: Channel) -> None:
    assert c.query("sip2sip.info", "NAPTR").result()


def test_query_ptr(c: Channel) -> None:
    assert c.query(
        ipaddress.ip_address("172.253.122.26").reverse_pointer, "PTR"
    ).result()


def test_query_bad_type(c: Channel) -> None:
    with pytest.raises(ValueError):
        c.query("google.com", "XXX")


def test_query_bad_class(c: Channel) -> None:
    with pytest.raises(TypeError):
        c.query("google.com", "A", query_class="INVALIDCLASS").result()


@pytest.mark.skipif(sys.platform == "darwin", reason="hangs")
def test_mx_dns_search(c: Channel) -> None:
    fut = c.search("gmail.com", query_type="MX").result()
    assert any([mx.host == b"gmail-smtp-in.l.google.com" for mx in fut])


def test_a_dns_search(c: Channel) -> None:
    for f in [
        c.search("google.com", "A"),
        c.search("llhttp.org", "A"),
        c.search("llparse.org", "A"),
    ]:
        assert f.result()


def test_search_aaaa(c: Channel) -> None:
    assert c.search("ipv6.google.com", "AAAA").result()


def test_search_cname(c: Channel) -> None:
    assert c.search("www.amazon.com", "CNAME").result()


def test_search_mx(c: Channel) -> None:
    assert c.search("gmail.com", "MX").result()


def test_search_ns(c: Channel) -> None:
    assert c.search("google.com", "NS").result()


def test_search_txt(c: Channel) -> None:
    assert c.search("google.com", "TXT").result()


def test_search_soa(c: Channel) -> None:
    assert c.search("google.com", "SOA").result()


# I'll mix in bytes to speedup the remaining tests incase of hanging
def test_search_srv(c: Channel) -> None:
    assert c.search(b"_xmpp-server._tcp.jabber.org", "SRV").result()


def test_search_naptr(c: Channel) -> None:
    assert c.search(b"sip2sip.info", "NAPTR").result()


def test_search_ptr(c: Channel) -> None:
    assert c.search(
        ipaddress.ip_address("172.253.122.26").reverse_pointer, "PTR"
    ).result()


# TODO: Test getsock and a few other missing functions in a future update.


def test_gethostbyname(c: Channel) -> None:
    # Lets change hosts up a notch...
    assert c.gethostbyname(b"python.org", socket.AF_INET).result()



"""
aio
---

A Version of Aiodns if Aiodns theoretically supported cyares by default.
This module is experimental and could be removed in the future if merged
with aiodns.

"""

# Inspired by aiodns, joint library/modue was made incase
# aiodns devs decided not to adopt this library optionally.
# This was also made to test socket callbacks to see if they
# were working properly...
from __future__ import annotations

import asyncio
import socket
import sys
from concurrent.futures import Future as cc_Future
from logging import getLogger
from typing import Any, Iterable, Literal, Sequence, TypeVar, overload

from deprecated_params import deprecated_params

from .channel import *  # type: ignore
from .exception import AresError
from .resulttypes import *

_T = TypeVar("_T")


WINDOWS_SELECTOR_ERR_MSG = (
    "cyares.aio cannot use ProactorEventLoop on Windows"
    " if cyares has no threadsafety. See more: "
    "https://github.com/aio-libs/aiodns#note-for-windows-users",
)

_LOGGER = getLogger(__name__)


query_type_map = {
    "A": QUERY_TYPE_A,
    "AAAA": QUERY_TYPE_AAAA,
    "ANY": QUERY_TYPE_ANY,
    "CAA": QUERY_TYPE_CAA,
    "CNAME": QUERY_TYPE_CNAME,
    "MX": QUERY_TYPE_MX,
    "NAPTR": QUERY_TYPE_NAPTR,
    "NS": QUERY_TYPE_NS,
    "PTR": QUERY_TYPE_PTR,
    "SOA": QUERY_TYPE_SOA,
    "SRV": QUERY_TYPE_SRV,
    "TXT": QUERY_TYPE_TXT,
}

query_class_map = {
    "IN": QUERY_CLASS_IN,
    "CHAOS": QUERY_CLASS_CHAOS,
    "HS": QUERY_CLASS_HS,
    "NONE": QUERY_CLASS_NONE,
    "ANY": QUERY_CLASS_ANY,
}


@deprecated_params(
    ["sock_state_cb"],
    {
        "socket_state_cb": "providing socket_state_cb will throw an Exception instead "
        "of being ignored in a future version of cyares"
    },
    removed_in="0.1.8",
)
class DNSResolver:
    """Simillar to aiodns's version but it aims to be more compact and have better typehints"""

    def __init__(
        self,
        servers: list[str] | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
        event_thread: bool = True,
        timeout: float | None = None,
        flags: int | None = None,
        tries: int | None = None,
        ndots: object | None = None,
        tcp_port: int | None = None,
        udp_port: int | None = None,
        domains: list[str] | None = None,
        lookups: str | bytes | bytearray | memoryview[int] | None = None,
        socket_send_buffer_size: int | None = None,
        socket_receive_buffer_size: int | None = None,
        rotate: bool = False,
        local_ip: str | bytes | bytearray | memoryview[int] | None = None,
        local_dev: str | bytes | bytearray | memoryview[int] | None = None,
        resolvconf_path=None,
        **kwargs,
    ) -> None:
        """
        Params
        ------

        :param servers: a list of dns servers to connect to
        :param loop: the asyncio event loop to utilize, supported
            eventloops include, winloop, uvloop and the asyncio standard library
            NOTE for windows users: `SelectorEventLoop` & `winloop.Loop` are the only
            eventloops you can use when the event_thread is disabled or when event_thread 
            is not supported SEE: https://github.com/aio-libs/aiodns#note-for-windows-users
            for more information
 
        :param event_thread: if enabled try to see if cyares can utilize \
            event threads otherwise fallback to using a socket callback handle \
            if `false` socket callback will be used no matter what,
            setting to `false` is good for testing purposes or 
            when trying to act threadsafe until closure.
            A quick word of caution , deletion is **Not Thread-Safe**
            SEE: https://c-ares.org/docs/ares_destroy.html

        """
        self._closed = True
        self.loop = loop or asyncio.get_event_loop()

        # XXX: Do not use, we override this argument by default
        kwargs.pop("sock_state_cb", None)
        self._timeout = timeout
        # Internal (Using with pytest to help debug socket_cb handles)
        if event_thread:
            # Still in the process of trying to convince aiodns maintainers about why
            # passing arguments in this format is a good idea
            self._event_thread, self._channel = self._make_channel(
                flags=flags,
                tries=tries,
                ndots=ndots,
                tcp_port=tcp_port,
                udp_port=udp_port,
                domains=domains,
                lookups=lookups,
                socket_send_buffer_size=socket_send_buffer_size,
                socket_receive_buffer_size=socket_receive_buffer_size,
                rotate=rotate,
                local_ip=local_ip,
                local_dev=local_dev,
                resolvconf_path=resolvconf_path,
                **kwargs,
            )

        else:
            self._raise_if_windows_proctor()
            self._event_thread, self._channel = (
                False,
                Channel(
                    sock_state_cb=self._sock_state_cb,
                    flags=flags,
                    tries=tries,
                    ndots=ndots,
                    tcp_port=tcp_port,
                    udp_port=udp_port,
                    domains=domains,
                    lookups=lookups,
                    socket_send_buffer_size=socket_send_buffer_size,
                    socket_receive_buffer_size=socket_receive_buffer_size,
                    rotate=rotate,
                    local_ip=local_ip,
                    local_dev=local_dev,
                    resolvconf_path=None,
                    **kwargs,
                ),
            )

        if servers:
            self.nameservers = servers
        self._read_fds: set[int] = set()
        self._write_fds: set[int] = set()
        self._timer: asyncio.TimerHandle | None = None
        self._closed = False

    def _raise_if_windows_proctor(self):
        if sys.platform == "win32" and type(self.loop) is asyncio.ProactorEventLoop:
            raise RuntimeError(WINDOWS_SELECTOR_ERR_MSG)

    def _wrap_future(self, fut: cc_Future[_T]) -> asyncio.Future[_T]:
        return asyncio.wrap_future(fut)

    def _make_channel(self, **kwargs: Any) -> tuple[bool, Channel]:
        if cyares_threadsafety():
            # CyAres is Threadsafe
            try:
                return True, Channel(event_thread=True, timeout=self._timeout, **kwargs)
            except AresError as e:
                if sys.platform == "linux":
                    _LOGGER.warning(
                        "Failed to create DNS resolver channel with automatic "
                        "monitoring of resolver configuration changes. This "
                        "usually means the system ran out of inotify watches. "
                        "Falling back to socket state callback. Consider "
                        "increasing the system inotify watch limit: %s",
                        e,
                    )
                else:
                    _LOGGER.warning(
                        "Failed to create DNS resolver channel with automatic "
                        "monitoring of resolver configuration changes. "
                        "Falling back to socket state callback: %s",
                        e,
                    )

        self._raise_if_windows_proctor()

        return False, Channel(timeout=self._timeout, sock_state_cb=self._sock_state_cb, **kwargs)

    def _timer_cb(self) -> None:
        if self._read_fds or self._write_fds:
            self._channel.process_fd(CYARES_SOCKET_BAD, CYARES_SOCKET_BAD)
            self._start_timer()
        else:
            self._timer = None

    def _start_timer(self) -> None:
        timeout = self._timeout
        if timeout is None or timeout < 0 or timeout > 1:
            timeout = 1
        elif timeout == 0:
            timeout = 0.1

        self._timer = self.loop.call_later(timeout, self._timer_cb)

    def cancel(self) -> None:
        """Cancels all running futures queued by this dns resolver"""
        self._channel.cancel()

    async def _cleanup(self) -> None:
        """Cleanup timers and file descriptors when closing resolver."""
        # TODO: Add Checkpoint so things remain asynchronous
        if self._closed:
            return
        # Mark as closed first to prevent double cleanup
        self._closed = True
        # Cancel timer if running
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None

        # Remove all file descriptors
        for fd in self._read_fds:
            self.loop.remove_reader(fd)
        for fd in self._write_fds:
            self.loop.remove_writer(fd)

        self._read_fds.clear()
        self._write_fds.clear()
        self.cancel()

    def _sock_state_cb(self, fd: int, readable: bool, writable: bool) -> None:
        if readable or writable:
            if readable:
                self.loop.add_reader(fd, self._channel.process_read_fd, fd)
                self._read_fds.add(fd)
            if writable:
                self.loop.add_writer(fd, self._channel.process_write_fd, fd)
                self._write_fds.add(fd)
            if self._timer is None:
                self._start_timer()
        else:
            # socket is now closed
            if fd in self._read_fds:
                self._read_fds.discard(fd)
                self.loop.remove_reader(fd)

            if fd in self._write_fds:
                self._write_fds.discard(fd)
                self.loop.remove_writer(fd)

            if not self._read_fds and not self._write_fds and self._timer is not None:
                self._timer.cancel()
                self._timer = None

    @property
    def nameservers(self) -> Sequence[str]:
        return self._channel.servers

    @nameservers.setter
    def nameservers(self, value: Iterable[str | bytes]) -> None:
        self._channel.servers = value

    @overload
    def query(
        self, host: str, qtype: Literal["A"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_a_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["AAAA"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_aaaa_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["CAA"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_caa_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["CNAME"], qclass: str | None = ...
    ) -> asyncio.Future[ares_query_cname_result]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["MX"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_mx_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["NAPTR"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_naptr_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["NS"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_ns_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["PTR"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_ptr_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["SOA"], qclass: str | None = ...
    ) -> asyncio.Future[ares_query_soa_result]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["SRV"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_srv_result]]: ...
    @overload
    def query(
        self, host: str, qtype: Literal["TXT"], qclass: str | None = ...
    ) -> asyncio.Future[list[ares_query_txt_result]]: ...

    def query(
        self, host: str, qtype: str, qclass: str | None = None
    ) -> asyncio.Future[list[Any]] | asyncio.Future[Any]:
        try:
            qtype = query_type_map[qtype]
        except KeyError as e:
            raise ValueError(f"invalid query type: {qtype}") from e
        if qclass is not None:
            try:
                qclass = query_class_map[qclass]
            except KeyError as e:
                raise ValueError(f"invalid query class: {qclass}") from e

        # we use a different technique than pycares to try and
        # aggressively prevent vulnerabilities

        return self._wrap_future(self._channel.query(host, qtype, qclass))

    def gethostbyname(
        self, host: str, family: socket.AddressFamily
    ) -> asyncio.Future[ares_host_result]:
        return self._wrap_future(self._channel.gethostbyname(host, family))

    def getaddrinfo(
        self,
        host: str,
        family: socket.AddressFamily = socket.AF_UNSPEC,
        port: int | None = None,
        proto: int = 0,
        type: int = 0,
        flags: int = 0,
    ) -> asyncio.Future[ares_addrinfo_result]:
        return self._wrap_future(
            self._channel.getaddrinfo(
                host, port, family=family, type=type, proto=proto, flags=flags
            )
        )

    def gethostbyaddr(
        self, name: str | bytes | bytearray | memoryview[int]
    ) -> asyncio.Future[ares_host_result]:
        return self._wrap_future(self._channel.gethostbyaddr(name))

    async def close(self) -> None:
        """
        Cleanly close the DNS resolver.

        This should be called to ensure all resources are properly released.
        After calling close(), the resolver should not be used again.
        """
        await self._cleanup()

    def getnameinfo(
        self,
        sockaddr: tuple[str, int] | tuple[str, int, int, int],
        flags: int = 0,
    ) -> asyncio.Future[ares_nameinfo_result]:
        return self._wrap_future(self._channel.getnameinfo(sockaddr, flags))

    # Still needs a little bit more work on...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["A"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_a_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["AAAA"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_aaaa_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["CAA"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_caa_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["CNAME"], qclass: str | None = ...
    # ) -> asyncio.Future[ares_query_cname_result]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["MX"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_mx_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["NAPTR"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_naptr_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["NS"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_ns_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["PTR"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_ptr_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["SOA"], qclass: str | None = ...
    # ) -> asyncio.Future[ares_query_soa_result]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["SRV"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_srv_result]]: ...
    # @overload
    # def search(
    #     self, host: str, qtype: Literal["TXT"], qclass: str | None = ...
    # ) -> asyncio.Future[list[ares_query_txt_result]]: ...

    # def search(
    #     self, host: str, qtype: str, qclass: str | None = None
    # ) -> asyncio.Future[list[Any]] | asyncio.Future[Any]:
    #     try:
    #         qtype = query_type_map[qtype]
    #     except KeyError as e:
    #         raise ValueError(f"invalid query type: {qtype}") from e
    #     if qclass is not None:
    #         try:
    #             qclass = query_class_map[qclass]
    #         except KeyError as e:
    #             raise ValueError(f"invalid query class: {qclass}") from e

    #     return self._wrap_future(self._channel.query(host, qtype, qclass))

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()

    # TODO: I will do the rest of the functionality a bit later...



from socket import htons, htonl
from cyares.channel import __htons, __htonl  # type: ignore


def test_cyares_htons() -> None:
    x = __htons(0xF4F5)
    assert hex(x) == hex(htons(0xF4F5))
    assert __htons(0xFEED) == htons(0xFEED)
    assert __htons(0xBEEF) == htons(0xBEEF)
    assert __htons(0x1166) == htons(0x1166)


def test_cyares_htonl() -> None:
    assert __htonl(0xFEEDBEEF) == htonl(0xFEEDBEEF)
    assert __htonl(0x11663322) == htonl(0x11663322)



from typing import Optional, Union

__test__: dict

class AresResult:
    def __repr__(self) -> str: ...
    @property
    def type() -> str: ...

class ares_optval_result(AresResult):
    val: str
    id: int

class ares_query_hinfo_result(AresResult):
    cpu: bytes
    os: bytes

class ares_query_sig_result(AresResult):
    type_covered: int
    algorithm: int
    labels: int
    original_ttl: int
    expiration: int
    inception: int
    key_tag: int
    signature: str
    signers_name: bytes

class ares_query_tlsa_result(AresResult):
    cert_usage: int
    selector: int
    data: str

    # XXX: Might be syntax problems due to match introduced in 3.10
    # so here's a workaround...
    @property
    def match() -> int: ...

class ares_query_svcb_result(AresResult):
    priority: int
    target: bytes
    params: list[ares_optval_result]

class ares_query_https_result(AresResult):
    priority: int
    target: bytes
    params: list[ares_optval_result]

class ares_query_uri_result(AresResult):
    priority: int
    weight: int
    target: bytes

class ares_query_raw_rr_result(AresResult):
    ty: int
    data: str

class ares_addrinfo_cname_result(AresResult):
    alias: bytes
    name: bytes
    ttl: int

class ares_addrinfo_node_result(AresResult):
    # One of two things is (address, port) or adding in flow_info and scope_id
    addr: Union[tuple[bytes, int], tuple[bytes, int, int, int]]
    family: int
    flags: int
    protocol: int
    socktype: int
    ttl: int

class ares_addrinfo_result(AresResult):
    cnames: list[ares_addrinfo_cname_result]
    nodes: list[ares_addrinfo_node_result]

class ares_host_result(AresResult):
    addresses: list[bytes]
    aliases: list[bytes]
    name: bytes

class ares_nameinfo_result(AresResult):
    node: bytes
    service: Optional[bytes]

class ares_query_a_result(AresResult):
    host: bytes
    ttl: int

class ares_query_aaaa_result(AresResult):
    host: bytes
    ttl: int

class ares_query_caa_result(AresResult):
    critical: int
    property: bytes
    ttl: int
    value: bytes

class ares_query_cname_result(AresResult):
    cname: bytes
    ttl: int

class ares_query_mx_result(AresResult):
    host: bytes
    priority: int
    ttl: int

class ares_query_naptr_result(AresResult):
    host: bytes
    ttl: int

class ares_query_ns_result(AresResult):
    host: bytes

class ares_query_ptr_result(AresResult):
    aliases: list[bytes]
    name: bytes
    ttl: int

class ares_query_soa_result(AresResult):
    nsname: bytes
    hostmaster: bytes
    serial: int
    referesh: int
    retry: int
    expire: int
    minttl: int
    ttl: int

class ares_query_srv_result(AresResult):
    host: bytes
    port: int
    priority: int

class ares_query_txt_result(AresResult):
    text: bytes
    ttl: int



"""
Exceptions
----------

This only contains one class and that is AresError
"""

class AresError(Exception):
    """Raised when c-ares fails to do something"""

    status: int
    strerror: bytes
    name: bytes
    def __init__(self, status: int) -> None: ...
    def __str__(self) -> str: ...



"""
aiohttp
-------

DNS Resolver Extension for replacing `AsyncResolver` or `ThreadedResolver`
in trade for `cyares` for handling aiohttp related http requests.

## Using Globally (Monkey Patch Method)

`install()` applies a monkeypatch globally
Know that if you wanted to use different resolvers with
aiohttp you can do that by passing `CyAresResolver` through to the
`ClientSession`. another way
to avoid `install()` not being threadsafe is by
installing before the rest of your code begins running.
Same as how **uvloop** and **winloop** both work

::

    from cyares.aiohttp import install

    if __name__ == "__main__":
        install()

when your done with an evenloop run and want to rerun
another but with an alternate dnsresolver such as **aiodns**
or the `ThreadedResolver` use `uninstall()`

# Using Non-Globally

Although slightly more tricky it is possible to pass along `CyAresResolver`
with the right setup without resorting to `install()`. This will also
work with aiohttp_socks or using aiohttp_socks with a proxy or tor or i2p.
If you feel uncomfortable with sending this much stuff to `ClientSession`
yourself it's encouraged to use globally with `install()` or via
monkeypatching aiohttp's default dns resolver in your own way.

::

    from cyares.aiohttp import CyAresResolver
    from aiohttp import ClientSession, TCPConnector

    # NOTE: This should also be respected when aiohttp_socks is in use
    # Such as the ProxyConnector which is another connector type of it's
    # own

    async def main():
        async with ClientSession(
            connector=TCPConnector(CyAresResolver()
        )) as client:
            async with client.get("https://httpbin.org/ip") as result:
                data = await result.json()
                print(f'YOUR IP ADDRESS IS: {data["origin"]}')

"""

import asyncio
import socket
from typing import Optional, Union

import aiohttp
from aiohttp.abc import AbstractResolver, ResolveResult

from .aio import DNSResolver
from .exception import AresError

_NUMERIC_SOCKET_FLAGS = socket.AI_NUMERICHOST | socket.AI_NUMERICSERV
_AI_ADDRCONFIG = socket.AI_ADDRCONFIG
if hasattr(socket, "AI_MASK"):
    _AI_ADDRCONFIG &= socket.AI_MASK


class CyAresResolver(AbstractResolver):
    def __init__(
        self, loop: Optional[asyncio.AbstractEventLoop] = None, *args, **kwargs
    ) -> None:
        # we already have getaddrinfo implemented in our version so
        # no need to write more functions than the required ones
        self._resolver = DNSResolver(loop=loop, *args, **kwargs)
        self._loop = loop or asyncio.get_event_loop()

    async def resolve(
        self, host: str, port: int = 0, family=socket.AF_INET
    ) -> list[ResolveResult]:
        """Return IP address for given hostname"""
        try:
            resp = await self._resolver.getaddrinfo(
                host=host,
                port=port,
                type=socket.SOCK_STREAM,
                family=family,
                flags=_AI_ADDRCONFIG,
            )
        except AresError as exc:
            msg = (
                exc.strerror.decode("utf-8", "surrogateescape")
                if exc.status >= 1
                else "DNS lookup failed"
            )
            raise OSError(None, msg) from exc
        hosts: list[ResolveResult] = []
        for node in resp.nodes:
            address: Union[tuple[bytes, int], tuple[bytes, int, int, int]] = node.addr
            family = node.family
            if family == socket.AF_INET6:
                # check scope ID to see if we need
                # to recurse
                if len(address) > 3 and address[3]:
                    result = await self._resolver.getnameinfo(
                        (address[0].decode("ascii"), *address[1:])
                    )
                    # XXX: Seems aiohttp forgot about
                    # decoding ascii here Maybe a pull request on their end would fix that?
                    resolved_host = result.node.decode("ascii")
                else:
                    resolved_host = address[0].decode("ascii")
                    port = address[1]
            else:
                assert family == socket.AF_INET
                resolved_host = address[0].decode("ascii")
                port = address[1]

            hosts.append(
                ResolveResult(
                    hostname=host,
                    host=resolved_host,
                    port=port,
                    family=family,
                    proto=0,
                    flags=_NUMERIC_SOCKET_FLAGS,
                )
            )

        # There were no results
        if not hosts:
            raise OSError(None, "DNS lookup failed")

        return hosts

    async def close(self) -> None:
        # Our version of aiodns has a different mechanism for closure
        # we need all these to close so that cleanup works correctly.
        await self._resolver.close()


PreviousDefaultResolver = aiohttp.resolver.DefaultResolver
"""Module Variable that is responsible for undoing CyAres 
aiohttp installation if nessesary"""


def install():
    """Simillar to how winloop & uvloop work, this function
    Monkey-Patches aiohttp's `DefaultResolver` and replaces it
    in trade for the `CyAresResolver`"""
    aiohttp.resolver.DefaultResolver = CyAresResolver


def uninstall():
    """Replaces Monkey-Patched aiohttp resolver for what was there originally
    good for situations such as testing different DNS Resolvers with `pytest`"""
    aiohttp.resolver.DefaultResolver = PreviousDefaultResolver



# NOTE: many parts were borrowed from pycares so that we wouldn't need to reinvent the wheel.

import os
import sys
import select

from setuptools.command.build_ext import build_ext
from setuptools import setup, Extension

use_system_lib = bool(int(os.environ.get("CYARES_USE_SYSTEM_LIB", 0)))

cares_sources = [
    "deps/c-ares/src/lib/ares_addrinfo_localhost.c",
    "deps/c-ares/src/lib/ares_addrinfo2hostent.c",
    "deps/c-ares/src/lib/ares_android.c",
    "deps/c-ares/src/lib/ares_cancel.c",
    "deps/c-ares/src/lib/ares_close_sockets.c",
    "deps/c-ares/src/lib/ares_conn.c",
    "deps/c-ares/src/lib/ares_cookie.c",
    "deps/c-ares/src/lib/ares_data.c",
    "deps/c-ares/src/lib/ares_destroy.c",
    "deps/c-ares/src/lib/ares_free_hostent.c",
    "deps/c-ares/src/lib/ares_free_string.c",
    "deps/c-ares/src/lib/ares_freeaddrinfo.c",
    "deps/c-ares/src/lib/ares_getaddrinfo.c",
    "deps/c-ares/src/lib/ares_getenv.c",
    "deps/c-ares/src/lib/ares_gethostbyaddr.c",
    "deps/c-ares/src/lib/ares_gethostbyname.c",
    "deps/c-ares/src/lib/ares_getnameinfo.c",
    "deps/c-ares/src/lib/ares_hosts_file.c",
    "deps/c-ares/src/lib/ares_init.c",
    "deps/c-ares/src/lib/ares_library_init.c",
    "deps/c-ares/src/lib/ares_metrics.c",
    "deps/c-ares/src/lib/ares_options.c",
    "deps/c-ares/src/lib/ares_parse_into_addrinfo.c",
    "deps/c-ares/src/lib/ares_process.c",
    "deps/c-ares/src/lib/ares_qcache.c",
    "deps/c-ares/src/lib/ares_query.c",
    "deps/c-ares/src/lib/ares_search.c",
    "deps/c-ares/src/lib/ares_send.c",
    "deps/c-ares/src/lib/ares_set_socket_functions.c",
    "deps/c-ares/src/lib/ares_socket.c",
    "deps/c-ares/src/lib/ares_sortaddrinfo.c",
    "deps/c-ares/src/lib/ares_strerror.c",
    "deps/c-ares/src/lib/ares_sysconfig_files.c",
    "deps/c-ares/src/lib/ares_sysconfig.c",
    "deps/c-ares/src/lib/ares_timeout.c",
    "deps/c-ares/src/lib/ares_update_servers.c",
    "deps/c-ares/src/lib/ares_version.c",
    "deps/c-ares/src/lib/dsa/ares_array.c",
    "deps/c-ares/src/lib/dsa/ares_htable_asvp.c",
    "deps/c-ares/src/lib/dsa/ares_htable_dict.c",
    "deps/c-ares/src/lib/dsa/ares_htable_strvp.c",
    "deps/c-ares/src/lib/dsa/ares_htable_szvp.c",
    "deps/c-ares/src/lib/dsa/ares_htable_vpstr.c",
    "deps/c-ares/src/lib/dsa/ares_htable_vpvp.c",
    "deps/c-ares/src/lib/dsa/ares_htable.c",
    "deps/c-ares/src/lib/dsa/ares_llist.c",
    "deps/c-ares/src/lib/dsa/ares_slist.c",
    "deps/c-ares/src/lib/event/ares_event_configchg.c",
    "deps/c-ares/src/lib/event/ares_event_thread.c",
    "deps/c-ares/src/lib/event/ares_event_wake_pipe.c",
    "deps/c-ares/src/lib/inet_net_pton.c",
    "deps/c-ares/src/lib/inet_ntop.c",
    "deps/c-ares/src/lib/legacy/ares_create_query.c",
    "deps/c-ares/src/lib/legacy/ares_expand_name.c",
    "deps/c-ares/src/lib/legacy/ares_expand_string.c",
    "deps/c-ares/src/lib/legacy/ares_fds.c",
    "deps/c-ares/src/lib/legacy/ares_getsock.c",
    "deps/c-ares/src/lib/legacy/ares_parse_a_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_aaaa_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_caa_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_mx_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_naptr_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_ns_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_ptr_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_soa_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_srv_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_txt_reply.c",
    "deps/c-ares/src/lib/legacy/ares_parse_uri_reply.c",
    "deps/c-ares/src/lib/record/ares_dns_mapping.c",
    "deps/c-ares/src/lib/record/ares_dns_multistring.c",
    "deps/c-ares/src/lib/record/ares_dns_name.c",
    "deps/c-ares/src/lib/record/ares_dns_parse.c",
    "deps/c-ares/src/lib/record/ares_dns_record.c",
    "deps/c-ares/src/lib/record/ares_dns_write.c",
    "deps/c-ares/src/lib/str/ares_buf.c",
    "deps/c-ares/src/lib/str/ares_str.c",
    "deps/c-ares/src/lib/str/ares_strsplit.c",
    "deps/c-ares/src/lib/util/ares_iface_ips.c",
    "deps/c-ares/src/lib/util/ares_math.c",
    "deps/c-ares/src/lib/util/ares_rand.c",
    "deps/c-ares/src/lib/util/ares_threads.c",
    "deps/c-ares/src/lib/util/ares_timeval.c",
    "deps/c-ares/src/lib/util/ares_uri.c",
]

if sys.platform == "win32":
    cares_sources += [
        "deps/c-ares/src/lib/ares_sysconfig_win.c",
        "deps/c-ares/src/lib/windows_port.c",
        "deps/c-ares/src/lib/event/ares_event_win32.c",
    ]

if sys.platform == "darwin":
    cares_sources += ["deps/c-ares/src/lib/ares_sysconfig_mac.c"]


class cares_build_ext(build_ext):
    cares_dir = os.path.join("deps", "c-ares")
    build_config_dir = os.path.join("deps", "build-config")
    ext_headers = os.path.join("cyares")

    # Brought over from winloop since these can be very useful.
    user_options = build_ext.user_options + [
        ("cython-always", None, "run cythonize() even if .c files are present"),
        (
            "cython-annotate",
            None,
            "Produce a colorized HTML version of the Cython source.",
        ),
        ("cython-directives=", None, "Cythion compiler directives"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.cython_always = False
        self.cython_annotate = False
        self.cython_directives = None
        
    def add_include_dir(self, dir, force=False):
        if use_system_lib and not force:
            return
        dirs = self.compiler.include_dirs
        dirs.insert(0, dir)
        self.compiler.set_include_dirs(dirs)

    def build_extensions(self):
        self.add_include_dir(self.ext_headers)
        # make sure we get the inc folder
        self.add_include_dir(os.path.join(self.ext_headers, "inc"))
        self.add_include_dir(os.path.join(self.cares_dir, "include"))
        self.add_include_dir(os.path.join(self.cares_dir, "src", "lib"))
        self.add_include_dir(os.path.join(self.cares_dir, "src", "lib", "include"))
        self.add_include_dir(os.path.join(self.build_config_dir, "include"), True)
        if sys.platform != "win32":
            self.compiler.define_macro("HAVE_CONFIG_H", 1)
            self.compiler.define_macro("_LARGEFILE_SOURCE", 1)
            self.compiler.define_macro("_FILE_OFFSET_BITS", 64)
        if sys.platform.startswith("linux"):
            # Check if it's actually Android
            if os.environ.get("ANDROID_ROOT") and os.environ.get("ANDROID_DATA"):
                self.add_include_dir(
                    os.path.join(self.build_config_dir, "config_android")
                )
            else:
                self.add_include_dir(
                    os.path.join(self.build_config_dir, "config_linux")
                )
            self.compiler.add_library("dl")
            self.compiler.add_library("rt")
        elif sys.platform == "darwin":
            self.add_include_dir(os.path.join(self.build_config_dir, "config_darwin"))
            self.compiler.define_macro("_DARWIN_USE_64_BIT_INODE", 1)
        elif sys.platform.startswith("freebsd"):
            self.add_include_dir(os.path.join(self.build_config_dir, "config_freebsd"))
            self.compiler.add_library("kvm")
        elif sys.platform.startswith("dragonfly"):
            self.add_include_dir(os.path.join(self.build_config_dir, "config_freebsd"))
        elif sys.platform.startswith("netbsd"):
            self.add_include_dir(os.path.join(self.build_config_dir, "config_netbsd"))
        elif sys.platform.startswith("openbsd"):
            self.add_include_dir(os.path.join(self.build_config_dir, "config_openbsd"))
        elif sys.platform.startswith("sunos"):
            self.add_include_dir(os.path.join(self.build_config_dir, "config_sunos"))
            self.compiler.add_library("socket")
            self.compiler.add_library("nsl")
            self.compiler.add_library("kstat")
        elif sys.platform == "cygwin":
            self.add_include_dir(os.path.join(self.build_config_dir, "config_cygwin"))
        elif sys.platform == "win32":
            if "mingw" not in self.compiler.compiler_type:
                self.extensions[0].extra_link_args = ["/NODEFAULTLIB:libcmt"]
            self.compiler.add_library("advapi32")
            self.compiler.add_library("iphlpapi")
            self.compiler.add_library("psapi")
            self.compiler.add_library("ws2_32")
            self.compiler.define_macro("CARES_HAVE_WINSOCK2_H", 1)
            self.compiler.define_macro("CARES_HAVE_WS2TCPIP_H", 1)
            self.compiler.define_macro("CARES_HAVE_WINDOWS_H", 1)

        if use_system_lib:
            self.compiler.add_library("cares")
        else:
            sources = cares_sources.copy()
            self.compiler.define_macro("CARES_THREADS", 1)
            self.compiler.define_macro("CARES_STATICLIB", 1)
            if hasattr(select, "poll"):
                sources += ["deps/c-ares/src/lib/event/ares_event_poll.c"]
                self.compiler.define_macro("HAVE_POLL", 1)
                self.compiler.define_macro("HAVE_POLL_H", 1)
            if hasattr(select, "kqueue"):
                sources += ["deps/c-ares/src/lib/event/ares_event_kqueue.c"]
                self.compiler.define_macro("HAVE_KQUEUE", 1)
                self.compiler.define_macro("HAVE_SYS_TYPES_H", 1)
                self.compiler.define_macro("HAVE_SYS_EVENT_H", 1)
                self.compiler.define_macro("HAVE_SYS_TIME_H", 1)
                self.compiler.define_macro("HAVE_FCNTL_H", 1)
            if hasattr(select, "epoll"):
                sources += ["deps/c-ares/src/lib/event/ares_event_epoll.c"]
                self.compiler.define_macro("HAVE_EPOLL", 1)
                self.compiler.define_macro("HAVE_SYS_EPOLL_H", 1)
                self.compiler.define_macro("HAVE_FCNTL_H", 1)
            if hasattr(os, "pipe") and sys.platform != "win32":
                sources += ["deps/c-ares/src/lib/event/ares_event_select.c"]
                self.compiler.define_macro("HAVE_PIPE", 1)

            # Currently all extensions need these sources...
            # In the future we may decide to compile the c-ares library
            # first then add it in after due to how long it can take to
            # compile everything

            # TODO: Optionally allow source compiling 
            # for now we have to get hacky to cut the time needed compiling this 
            # stuff by 5x times

            # for e in self.extensions:
            #     e.sources += sources


            c = self.compiler
            if not os.path.exists("build"):
                os.mkdir("build")
            objects = c.compile(sources, "build")
            c.create_static_lib(objects, output_libname="cares", output_dir="build")
            for e in self.extensions:
                if "build/cares" not in e.libraries:
                    e.libraries.append("build/cares") 
        super().build_extensions()

    # Copied from winloop
    def finalize_options(self):
        need_cythonize = self.cython_always
        cfiles = {}

        for extension in self.distribution.ext_modules:
            for i, sfile in enumerate(extension.sources):
                if sfile.endswith(".pyx"):
                    prefix, ext = os.path.splitext(sfile)
                    cfile = prefix + ".c"

                    if os.path.exists(cfile) and not self.cython_always:
                        extension.sources[i] = cfile
                    else:
                        if os.path.exists(cfile):
                            cfiles[cfile] = os.path.getmtime(cfile)
                        else:
                            cfiles[cfile] = 0
                        need_cythonize = True

        if need_cythonize:
            # import pkg_resources

            # Double check Cython presence in case setup_requires
            # didn't go into effect (most likely because someone
            # imported Cython before setup_requires injected the
            # correct egg into sys.path.
            try:
                import Cython
            except ImportError:
                raise RuntimeError(
                    "please install cython to compile cyares from source"
                )

            from Cython.Build import cythonize

            directives = {}
            if self.cython_directives:
                for directive in self.cython_directives.split(","):
                    k, _, v = directive.partition("=")
                    if v.lower() == "false":
                        v = False
                    if v.lower() == "true":
                        v = True
                    directives[k] = v
                self.cython_directives = directives

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                compiler_directives=directives,
                annotate=self.cython_annotate,
                emit_linenums=self.debug,
            )

        return super().finalize_options()


if __name__ == "__main__":
    setup(
        ext_modules=[
            Extension(
                "cyares.callbacks", ["cyares/callbacks.pyx"], extra_compile_args=["-O2"]
            ),
            Extension(
                "cyares.channel", ["cyares/channel.pyx"], extra_compile_args=["-O2"]
            ),
            Extension(
                "cyares.exception", ["cyares/exception.pyx"], extra_compile_args=["-O2"]
            ),
            Extension(
                "cyares.resulttypes",
                ["cyares/resulttypes.pyx"],
                extra_compile_args=["-O2"],
            ),
            Extension(
                "cyares.socket_handle",
                ["cyares/socket_handle.pyx"],
                extra_compile_args=["-O2"],
            ),
        ],
        cmdclass={"build_ext": cares_build_ext},
    )



import sys
from concurrent.futures import Future
from typing import Callable, Literal, TypeVar, overload

from .resulttypes import *

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

CYARES_SOCKET_BAD: int = ...

# Query types
QUERY_TYPE_A: int = 1
QUERY_TYPE_AAAA: int = 28
QUERY_TYPE_ANY: int = 255
QUERY_TYPE_CAA: int = 257
QUERY_TYPE_CNAME: int = 5
QUERY_TYPE_MX: int = 15
QUERY_TYPE_NAPTR: int = 35
QUERY_TYPE_NS: int = 2
QUERY_TYPE_PTR: int = 12
QUERY_TYPE_SOA: int = 6
QUERY_TYPE_SRV: int = 33
QUERY_TYPE_TXT: int = 16

# Query classes
QUERY_CLASS_IN: int = ...
QUERY_CLASS_CHAOS: int = ...
QUERY_CLASS_HS: int = ...
QUERY_CLASS_NONE: int = ...
QUERY_CLASS_ANY: int = ...

class Channel:
    event_thread: bool
    """Used for checking if event-thread is in use before using wait(...)
    This is good for when you plan to make event-thread optional in a user 
    program
    """

    def __init__(
        self,
        flags: int | None = None,
        timeout: int | float | None = None,
        tries: int | None = None,
        ndots: object | None = None,
        tcp_port: int | None = None,
        udp_port: int | None = None,
        servers: list[str] | None = None,
        domains: list[str] | None = None,
        lookups: str | bytes | bytearray | memoryview[int] | None = None,
        sock_state_cb: Callable[[int, bool, bool], None] = None,
        socket_send_buffer_size: int | None = None,
        socket_receive_buffer_size: int | None = None,
        rotate: bool = False,
        local_ip: str | bytes | bytearray | memoryview[int] | None = None,
        local_dev: str | bytes | bytearray | memoryview[int] | None = None,
        resolvconf_path=None,
        event_thread: bool = False,
    ) -> None: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["ANY", 255],
        callback: Callable[[Future[list[AresResult]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[AresResult]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["A", 1],
        callback: Callable[[Future[list[ares_query_a_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_a_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["AAAA", 28],
        callback: Callable[[Future[ares_query_aaaa_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_aaaa_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["CAA", 257],
        callback: Callable[[Future[list[ares_query_caa_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_caa_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["CNAME", 5],
        callback: Callable[[Future[ares_query_cname_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[ares_query_cname_result]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["MX", 15],
        callback: Callable[[Future[list[ares_query_mx_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_mx_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["NAPTR", 35],
        callback: Callable[[Future[ares_query_naptr_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_naptr_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["NS", 2],
        callback: Callable[[Future[ares_query_ns_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_ns_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["PTR", 12],
        callback: Callable[[Future[ares_query_ptr_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_ptr_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["SOA", 6],
        callback: Callable[[Future[ares_query_soa_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[ares_query_soa_result]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["SRV", 33],
        callback: Callable[[Future[ares_query_srv_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_srv_result]]: ...
    @overload
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["TXT", 16],
        callback: Callable[[Future[ares_query_txt_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_txt_result]]: ...
    def query(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: str | int,
        callback: Callable[[Future[AresResult]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[AresResult]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["A"],
        callback: Callable[[Future[ares_query_a_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_a_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["AAAA"],
        callback: Callable[[Future[ares_query_aaaa_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_aaaa_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["CAA"],
        callback: Callable[[Future[ares_query_caa_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_caa_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["CNAME"],
        callback: Callable[[Future[ares_query_cname_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[ares_query_cname_result]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["MX"],
        callback: Callable[[Future[list[ares_query_mx_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_mx_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["NAPTR"],
        callback: Callable[[Future[list[ares_query_naptr_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_naptr_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["NS"],
        callback: Callable[[Future[ares_query_ns_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_ns_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["PTR"],
        callback: Callable[[Future[list[ares_query_ptr_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_ptr_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["SOA"],
        callback: Callable[[Future[ares_query_soa_result]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[ares_query_soa_result]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["SRV"],
        callback: Callable[[Future[list[ares_query_srv_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_srv_result]]: ...
    @overload
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: Literal["TXT"],
        callback: Callable[[Future[list[ares_query_txt_result]]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[list[ares_query_txt_result]]: ...
    def search(
        self,
        name: str | bytes | bytearray | memoryview[int],
        query_type: str | int,
        callback: Callable[[Future[AresResult]], None] | None = ...,
        query_class: str | int | None = ...,
    ) -> Future[AresResult]: ...
    @property
    def servers(self) -> list[str]: ...
    @servers.setter
    def servers(self, servers: list[str]) -> None: ...
    def cancel(self) -> None: ...
    def reinit(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *args) -> None: ...
    def process_fd(self, read_fd: int, write_fd: int) -> None: ...

    # TODO (Vizonex) Pull request to pycares with the same new functions I made
    def process_read_fd(self, read_fd: int) -> None:
        """
        processes readable file-descriptor instead of needing to remember
        to set write-fd to CYARES_SOCKET_BAD

        Parameters
        ----------

        :param read_fd: the readable file descriptor
        """

    def process_write_fd(self, write_fd: int) -> None:
        """
        processes writable file-descriptor instead of needing to remember
        to set read-fd to CYARES_SOCKET_BAD

        Parameters
        ----------

        :param write_fd: the writeable file descriptor
        """

    def getaddrinfo(
        self,
        host: str | bytes | bytearray | memoryview[int],
        port: object = ...,
        callback: Callable[[Future[ares_addrinfo_result]], None] | None = ...,
        family: int = ...,
        socktype: int = ...,
        proto: int = ...,
        flags: int = ...,
    ) -> Future[ares_addrinfo_result]: ...
    def getnameinfo(
        self,
        address: tuple[str, int] | tuple[int, int, int, int],
        flags: int,
        callback: Callable[[Future[ares_nameinfo_result]], None] | None = ...,
    ) -> Future[ares_nameinfo_result]: ...
    def gethostbyname(
        self,
        name: str | bytes | bytearray | memoryview[int],
        family: int,
        callback: Callable[[Future[ares_nameinfo_result]], None] | None = ...,
    ) -> Future[ares_host_result]: ...
    def gethostbyaddr(
        self,
        name: str | bytes | bytearray | memoryview[int],
        family: int,
        callback: Callable[[Future[ares_nameinfo_result]], None] | None = ...,
    ) -> Future[ares_host_result]: ...
    def getsock(self) -> tuple[list[int], list[int]]: ...
    def set_local_dev(self, dev: str | bytes | bytearray | memoryview[int]) -> None: ...
    def set_local_ip(self, ip: str | bytes | bytearray | memoryview[int]) -> None: ...
    def close(self) -> None: ...
    @overload
    def timeout(self) -> float: ...
    def timeout(self, t: float = ...) -> float: ...
    def wait(self, timeout: float | int | None = None) -> bool:
        """Waits for all queries to close using `ares_queue_wait_emtpy`
        This function blocks until notified that the timeout expired or
        that all pending queries have been cancelled or completed.

        Parameters
        ----------

        :param timeout: A timeout in seconds as a float or integer object
            this object will be rounded to milliseconds, throws `TypeError`
            if object is not None or an `int` or `float` other wise it throws
            `ValueError` if the timeout is less than 0, default runs until
            all cancelled or closed

        :return: True on success, False if queries are still running
        :rtype bool:
        """

def cyares_threadsafety() -> bool:
    """
    pycares documentation says:
    Check if c-ares was compiled with thread safety support.

    :return: True if thread-safe, False otherwise.
    :rtype: bool
    """



import pytest
from cyares.trio import DNSResolver
from cyares.exception import AresError
import ipaddress
import anyio as anyio


# set the backend to just trio
@pytest.fixture()
def anyio_backend() -> str:
    return "trio"


@pytest.fixture(
    params=(True, False), ids=("trio-event-thread", "trio-socket-cb"), scope="function"
)
async def resolver(anyio_backend: str, request: pytest.FixtureRequest):
    # should be supported on all operating systems...

    async with DNSResolver(
        servers=[
            "1.0.0.1",
            "1.1.1.1",
            "141.1.27.249",
            "194.190.225.2",
            "194.225.16.5",
            "91.185.6.10",
            "194.2.0.50",
            "66.187.16.5",
            "83.222.161.130",
            "69.60.160.196",
            "194.150.118.3",
            "84.8.2.11",
            "195.175.39.40",
            "193.239.159.37",
            "205.152.6.20",
            "82.151.90.1",
            "144.76.202.253",
            "103.3.46.254",
            "5.144.17.119",
            "8.8.8.8",
            "8.8.4.4",
        ],
        event_thread=request.param,
        tries=3,
        timeout=10,
    ) as channel:
        yield channel


@pytest.mark.anyio
async def test_mx_dns_query(resolver: DNSResolver) -> None:
    assert await resolver.query("gmail.com", "MX")


@pytest.mark.anyio
async def test_a_dns_query(resolver: DNSResolver) -> None:
    assert await resolver.query("python.org", "A")


@pytest.mark.anyio
async def test_cancelling() -> None:
    async with DNSResolver(servers=["8.8.8.8", "8.8.4.4"]) as channel:
        for f in [
            channel.query("google.com", "A"),
            channel.query("llhttp.org", "A"),
            channel.query("llparse.org", "A"),
        ]:
            f.cancel()


@pytest.mark.anyio
async def test_cancelling_from_resolver() -> None:
    async with DNSResolver(servers=["8.8.8.8", "8.8.4.4"]) as resolver:
        _ = [
            resolver.query("google.com", "A"),
            resolver.query("llhttp.org", "A"),
            resolver.query("llparse.org", "A"),
        ]
        resolver.cancel()


@pytest.mark.anyio
async def test_a_dns_query_fail(resolver: DNSResolver) -> None:
    with pytest.raises(
        AresError,
        match=r"\[ARES_ENODATA : 1\] DNS server returned answer with no data",
    ):
        await resolver.query("hgf8g2od29hdohid.com", "A")


@pytest.mark.anyio
async def test_query_aaaa(resolver: DNSResolver) -> None:
    assert await resolver.query("ipv6.google.com", "AAAA")


@pytest.mark.anyio
async def test_query_cname(resolver: DNSResolver) -> None:
    assert await resolver.query("www.amazon.com", "CNAME")


@pytest.mark.anyio
async def test_query_mx(resolver: DNSResolver) -> None:
    assert await resolver.query("google.com", "MX")


@pytest.mark.anyio
async def test_query_ns(resolver: DNSResolver) -> None:
    assert await resolver.query("google.com", "NS")


@pytest.mark.anyio
async def test_query_txt(resolver: DNSResolver) -> None:
    assert await resolver.query("google.com", "TXT")


@pytest.mark.anyio
async def test_query_soa(resolver: DNSResolver) -> None:
    assert await resolver.query("google.com", "SOA")


@pytest.mark.anyio
async def test_query_srv(resolver: DNSResolver) -> None:
    assert await resolver.query("_xmpp-server._tcp.jabber.org", "SRV")


@pytest.mark.anyio
async def test_query_naptr(resolver: DNSResolver) -> None:
    assert await resolver.query("sip2sip.info", "NAPTR")


@pytest.mark.anyio
async def test_query_ptr(resolver: DNSResolver) -> None:
    assert await resolver.query(
        ipaddress.ip_address("172.253.122.26").reverse_pointer, "PTR"
    )


@pytest.mark.anyio
async def test_query_bad_type(resolver: DNSResolver) -> None:
    with pytest.raises(ValueError):
        await resolver.query("google.com", "XXX")


@pytest.mark.anyio
async def test_query_bad_class(resolver: DNSResolver) -> None:
    with pytest.raises(ValueError):
        await resolver.query("google.com", "A", qclass="INVALIDCLASS")
